unit test_voltage1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ExtCtrls, Grids,
  StdCtrls, TAGraph, TASeries, TATransformations;

type

  { TForm1 }

  TForm1 = class(TForm)
    Button2: TButton;
    btnExportCSV: TButton;
    btnImportPX4: TButton;
    btnImportSimple: TButton;
    Chart1: TChart;
    Chart1ConstantLine1: TConstantLine;
    Chart1ConstantLine2: TConstantLine;
    Chart1LineSeries1: TLineSeries;
    Chart1LineSeries2: TLineSeries;
    Chart1LineSeries4: TLineSeries;

    cgpCharts: TCheckGroup;
    Label1: TLabel;
    Label2: TLabel;
    OpenDialog1: TOpenDialog;
    rgNumCells: TRadioGroup;
    SaveDialog1: TSaveDialog;
    gridVoltage: TStringGrid;
    procedure Berechnen;
    procedure Button2Click(Sender: TObject);
    procedure btnExportCSVClick(Sender: TObject);
    procedure btnImportPX4Click(Sender: TObject);
    procedure btnImportSimpleClick(Sender: TObject);
    procedure cgpChartsItemClick(Sender: TObject; Index: integer);
    procedure FormCreate(Sender: TObject);
    procedure rgNumCellsClick(Sender: TObject);
    procedure gridVoltageAfterSelection(Sender: TObject; aCol, aRow: Integer);
  private

  public

  end;

var
  Form1: TForm1;

const
  lipomax=4.2;                                     {normal LiPo}
  appname='Test Voltage to Capacity';
  tab3='   ';
  frmout='0.0';
  capDefault='Cap [%] Y';
  capCSV='Cap [%] Ext';
  capPX4='Cap [%] PX4';

{The relationship of voltage and capacity from RC-Groups:
 https://blog.ampow.com/lipo-voltage-chart/
 https://www.rcgroups.com/forums/showpost.php?p=29431951
 https://www.flitetest.com/articles/Parallel_Charging_Your_LiPo_Batteries}

CapTab: array [0..20] of double = (
  100,  95,   90,   85,   80,   75,   70,   65,   60,   55,   50,
  45,   40,   35,   30,   25,   20,   15,   10,   5,    0);
s1tab: array [0..20] of double = (
  4.20, 4.15, 4.11, 4.08, 4.02, 3.98, 3.95, 3.91, 3.87, 3.85, 3.84,
  3.82, 3.80, 3.79, 3.77, 3.75, 3.73, 3.71, 3.69, 3.61, 3.27);
s2tab: array [0..20] of double = (
  8.40, 8.30, 8.22, 8.16, 8.05, 7.97, 7.91, 7.83, 7.75, 7.71, 7.67,
  7.63, 7.59, 7.57, 7.53, 7.49, 7.45, 7.41, 7.37, 7.22, 6.55);
s3tab: array [0..20] of double =(
  12.6, 12.45, 12.33, 12.25, 12.07, 11.95, 11.86, 11.74, 11.62, 11.56, 11.51,
  11.45, 11.39, 11.36, 11.3, 11.24, 11.18, 11.12, 11.06, 10.83, 9.82);
s4tab: array [0..20] of double = (
  16.8, 16.6,  16.45, 16.33, 16.09, 15.93, 15.81, 15.66, 15.5, 15.42, 15.34,
  15.26, 15.18, 15.14, 15.06, 14.99, 14.91, 14.83, 14.75, 14.43, 13.09);
s5tab: array [0..20] of double = (
  21.0, 20.75, 20.56, 20.41, 20.11, 19.92, 19.77, 19.57, 19.37, 19.27, 19.18,
  19.08, 18.98, 18.93, 18.83,18.73, 18.63, 18.54, 18.44, 18.04, 16.37);
s6tab: array [0..20] of double = (
  25.2, 24.9, 24.67, 24.49, 24.14, 23.9, 23.72, 23.48, 23.25, 23.13, 23.01,
  22.89, 22.77, 22.72, 22.6, 22.48, 22.36, 22.24, 22.12, 21.65, 19.64);

{PX4 Autopilot, genauer gesagt QGC:
  BAT1_V_CHARGED default: 4,05V
  BAT1_V_EMPTY   default: 3,5V
                 dazwischen linear von 100% bis 0% entspricht bei 4S 22mV per %
                 siehe Voltage_PX4QGC_simple.csv

}

implementation

{$R *.lfm}

{Voltage to capacity from RC-Groups table, smooth values like in Q500log2kml}

function VtoProzRC(const vt: integer; u: double): double;
var i: integer;
    uz: double;

begin
  result:=100;
  case vt of                                       {Number cells -1; Index of rgNumCells}
    0: uz:=u;                                      {1S}
    1: uz:=u/2;
    3: uz:=u/4;
    4: uz:=u/5;
    5: uz:=u/6;                                    {6S}
  else
    uz:=u/4;                                       {Default: 4S}
  end;
  if uz<S1Tab[high(S1Tab)] then result:=0
  else
  if uz<s1tab[0] then begin
    for i:=0 to high(CapTab) do begin
      if uz>S1Tab[i] then
        break;                                     {Voltage between index i-1 and i}
    end;
//    result:=round(CapTab[i]+((CapTab[i-1]-CapTab[i])/
//                            ((S1Tab[i-1]-S1Tab[i])/(uz-S1Tab[i]))));
    result:=CapTab[i]+((CapTab[i-1]-CapTab[i])/
                      ((S1Tab[i-1]-S1Tab[i])/(uz-S1Tab[i])));  {double}
  end;
end;

{from Yuneec ST1x source code: LiPo voltage to remaining capacity [%]}
function VtoProzY(const vt: integer; const u: double): integer; {# cells, voltage [V]}
const s61=23.9;                                    {Thresholds 6S}
      s62=21.7;
      s63=21.3;
      s64=21.1;

      s41=14.9;                                    {Thresholds 4S}
      s42=14.2;
      s43=14.0;
      s44=13.8;

      s31=10.7;                                    {Thresholds 3S}
      s32=10.5;
      s33=10.3;

      s21=7.2;                                     {Thresholds 2S my estimate, not Yuneec}
      s22=7.0;
      s23=6.8;

var   m: double;                                   {Max LipO voltage}
begin
  result:=0;                                       {default 0%}
  case vt of
     1: begin
          m:=2*lipomax;                            {my 2S}
          if  u>=m then
            result:=100;                           {100%}
          if (u>=s21) and
             (u< m)   then
            result:=round((((u-s21)*75)/(m-s21))+25);
          if (u> s22) and
             (u< s21) then
            result:=round((((u-s22)*20)/(s21-s22))+5);
          if (u> s23) and
             (u<=s22) then
            result:=round( ((u-s23)*5)/(s22-s23));
        end;
     2: begin
          m:=3*lipomax;                            {3S}
          if  u>=m then
            result:=100;                           {100%}
          if (u>=s31) and
             (u< m)   then
            result:=round((((u-s31)*75)/(m-  s31))+25);
          if (u> s32) and
             (u< s31) then
            result:=round((((u-s32)*20)/(s31-s32))+5);
          if (u> s33) and
             (u<=s32) then
            result:=round( ((u-s33)* 5)/(s32-s33));
         end;
     3: begin
          m:=4*lipomax;                            {4S}
          if  u>=m then
            result:=100;
          if (u>=s41) and
             (u< m)   then
            result:=round((((u-s41)*50)/(m-  s41))+50);
          if (u> s42) and
             (u< s41) then
            result:=round((((u-s42)*25)/(s41-s42))+25);
          if (u> s43) and
             (u< s42) then
            result:=round((((u-s43)*20)/(s42-s43))+5);
          if (u> s44) and
             (u<=s43) then
            result:=round( ((u-s44)* 5)/(s43-s44));
        end;
    5: begin
         m:=6*lipomax;                             {6S}
         if  u>=m then
            result:=100;
         if (u>=s61) and
            (u< m)   then
           result:=round((((u-s61)* 5)/(m-  s61))+95);
         if (u> s62) and
            (u< s61) then
           result:=round((((u-s62)*75)/(s61-s62))+20);
         if (u> s63) and
            (u< s62) then
           result:=round((((u-s63)* 5)/(s62-s63))+5);
         if (u> s64) and
            (u<=s63) then
           result:=round( ((u-s64)* 5)/(s63-s64));
       end;
  end;
end;

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
begin
  DefaultFormatSettings.DecimalSeparator:='.';
  gridVoltage.Cells[0, 0]:='#';
  gridVoltage.Cells[1, 0]:='Volts';
  gridVoltage.Cells[2, 0]:=capDefault;
  gridVoltage.Cells[3, 0]:='Cap [%] RC';
  btnExportCSV.Enabled:=false;
  Caption:=appname;
  cgpCharts.Checked[0]:=true;
  cgpCharts.Checked[1]:=true;
  Berechnen;
end;

procedure TForm1.rgNumCellsClick(Sender: TObject); {Run}
begin
  Chart1LineSeries4.Clear;                         {Clear CSV import chart}
  Berechnen;
end;

procedure TForm1.gridVoltageAfterSelection(Sender: TObject; aCol, aRow: Integer);
begin
  Chart1ConstantLine1.Position:=StrToFloat(gridVoltage.Cells[1, aRow]);
  Chart1ConstantLine2.Position:=StrToFloat(gridVoltage.Cells[2, aRow]);
end;

procedure TForm1.Berechnen;                        {Run computing}
var i: integer;
    p, p1, v: Double;
begin
  Chart1LineSeries1.Clear;                         {clear all charts: Yuneec ST1x}
  Chart1LineSeries2.Clear;                         {RC Groups table}
  gridVoltage.RowCount:=(4200-3200);               {4.2V down to 3.2V in 0.001 steps}
  Caption:=appname;
  v:=lipomax;

  for i:=1 to gridVoltage.RowCount-1 do begin
    gridVoltage.Cells[0, i]:=IntToStr(i);
    gridVoltage.Cells[1, i]:=FormatFloat('0.000', v*(rgNumCells.ItemIndex+1));
    p:=VtoProzY(rgNumCells.ItemIndex, v*(rgNumCells.ItemIndex+1));
    p1:=VtoProzRC(rgNumCells.ItemIndex, v*(rgNumCells.ItemIndex+1));
    gridVoltage.Cells[2, i]:=FormatFloat(frmout, p);           {Yuneec column}
    gridVoltage.Cells[3, i]:=FormatFloat(frmout, p1);          {RC-Group column}
    if cgpCharts.Checked[0] then
      Chart1LineSeries1.AddXY(p, v*(rgNumCells.ItemIndex+1));  {as Yuneec ST1x SW}
(*    if cgpCharts.Checked[1] then
      Chart1LineSeries2.AddXY(p1, v*(rgNumCells.ItemIndex+1)); {as Q500log2KML} *)
    v:=v-0.001;
  end;

  if cgpCharts.Checked[1] then begin
    for i:=0 to 20 do begin                        {RC Group table of fix values}
      case rgNumCells.ItemIndex of                 {Number cells}
        0: Chart1LineSeries2.AddXY(captab[i], s1tab[i]);
        1: Chart1LineSeries2.AddXY(captab[i], s2tab[i]);
        2: Chart1LineSeries2.AddXY(captab[i], s3tab[i]);
        3: Chart1LineSeries2.AddXY(captab[i], s4tab[i]);
        4: Chart1LineSeries2.AddXY(captab[i], s5tab[i]);
        5: Chart1LineSeries2.AddXY(captab[i], s6tab[i]);
      end;
    end;
  end;
  if gridVoltage.RowCount>1 then
    Chart1ConstantLine1.Position:=StrToFloat(gridVoltage.Cells[1, 1]);
end;

procedure TForm1.Button2Click(Sender: TObject);    {Close}
begin
  Close;
end;

procedure TForm1.btnExportCSVClick(Sender: TObject);    {Export to CSV}
begin
  SaveDialog1.FileName:='VtoProz_'+
                        rgNumCells.Items[rgNumCells.ItemIndex]+
                        '_1.csv';
  If SaveDialog1.Execute then begin
    gridVoltage.SaveToCSVFile(SaveDialog1.Filename, ';');
    btnExportCSV.Enabled:=false;
  end;
end;

{Q500log2kml output PX4 CSV files: Voltage and Remaining capacity from
 MAV_SYS_STATUS (1) - V has index 2 and Cap % has index 46}

procedure TForm1.btnImportPX4Click(Sender: TObject);    {load PX4 CSV file}
var inlist, splitlist: TStringList;
    i, z: integer;
    v: double;
begin
  if OpenDialog1.Execute then begin
    Screen.Cursor:=crHourGlass;
    inlist:=TStringlist.Create;
    splitlist:=TStringlist.Create;
    splitlist.Delimiter:=',';
    splitlist.StrictDelimiter:=true;
    try
      inlist.LoadFromFile(OpenDialog1.FileName);
      gridVoltage.BeginUpdate;
      if (inlist.Count>10) then begin
        gridVoltage.Cells[2, 0]:=capPX4;
        Caption:=appname+tab3+ExtractFileName(OpenDialog1.FileName);
        gridVoltage.RowCount:=1;                   {Clear table}
        Chart1LineSeries4.Clear;                   {Clear chart}
        gridVoltage.RowCount:=inlist.Count+1;
        z:=1;
        for i:=z to inlist.Count-1 do begin
          splitlist.DelimitedText:=inlist[i];
          if (splitlist[2]<>'') and
             (splitlist[46]<>'') then begin
            v:=StrToFloat(splitlist[2]);           {Voltage}
{Fill table just for info}
            gridVoltage.Cells[0, z]:=IntToStr(z);
            gridVoltage.Cells[1, z]:=splitlist[2];
            gridVoltage.Cells[2, z]:=splitlist[46];
            gridVoltage.Cells[3, z]:=
              FormatFloat(frmout, VtoProzRC(rgNumCells.ItemIndex, v));
{Create chart for PX4 CSV file}
            Chart1LineSeries4.AddXY(StrToInt(splitlist[46]), v);
            inc(z);
          end;
        end;
      end;
    finally
      gridVoltage.EndUpdate;
      inlist.Free;
      splitlist.Free;
      Screen.Cursor:=crDefault;
    end;
  end;
end;

{Simple, handmade CSV file: Voltage, Capacity (i.e. from GUI)
Example:
16.8,100
16.7,96.5
...}
procedure TForm1.btnImportSimpleClick(Sender: TObject);    {load simple CSV file}
var inlist, splitlist: TStringList;
    i, z: integer;
    v: double;
begin
  if OpenDialog1.Execute then begin
    Screen.Cursor:=crHourGlass;
    inlist:=TStringlist.Create;
    splitlist:=TStringlist.Create;
    splitlist.Delimiter:=',';
    splitlist.StrictDelimiter:=true;
    try
      inlist.LoadFromFile(OpenDialog1.FileName);
      gridVoltage.BeginUpdate;
      if (inlist.Count>4) then begin
        gridVoltage.Cells[2, 0]:=capCSV;
        Caption:=appname+tab3+ExtractFileName(OpenDialog1.FileName);
        gridVoltage.RowCount:=1;                   {Clear table}
        Chart1LineSeries4.Clear;                   {Clear chart}
        gridVoltage.RowCount:=inlist.Count+1;
        z:=1;
        for i:=z to inlist.Count-1 do begin
          splitlist.DelimitedText:=inlist[i];
          if (splitlist[0]<>'') and
             (splitlist[1]<>'') then begin
            v:=StrToFloat(splitlist[0]);           {Voltage}
{Fill table just for info}
            gridVoltage.Cells[0, z]:=IntToStr(z);
            gridVoltage.Cells[1, z]:=splitlist[0];
            gridVoltage.Cells[2, z]:=splitlist[1];
            gridVoltage.Cells[3, z]:=
              FormatFloat(frmout, VtoProzRC(rgNumCells.ItemIndex, v));
{Create chart for simple CSV file}
            Chart1LineSeries4.AddXY(StrToFloat(splitlist[1]), v);
            inc(z);
          end;
        end;
      end;
    finally
      gridVoltage.EndUpdate;
      inlist.Free;
      splitlist.Free;
      Screen.Cursor:=crDefault;
    end;
  end;
end;

procedure TForm1.cgpChartsItemClick(Sender: TObject; Index: integer);
begin                                              {Change selection of visible charts}
  Berechnen;
end;

end.

